#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Example property package for the saponification of Ethyl Acetate with NaOH
Assumes dilute solutions with properties of H2O.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    NonNegativeReals,
    Param,
    PositiveReals,
    Reals,
    units,
    value,
    Var,
)

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    LiquidPhase,
    Component,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
import idaes.logger as idaeslog
from idaes.core.scaling import CustomScalerBase

# Some more information about this module
__author__ = "Andrew Lee"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("SaponificationParameterBlock")
class PhysicalParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for
    superheated steam.

    """

    def build(self):
        """
        Callable method for Block construction.
        """
        super(PhysicalParameterData, self).build()

        self._state_block_class = SaponificationStateBlock

        # Add Phase objects
        self.Liq = LiquidPhase()

        # Add Component objects
        self.H2O = Component()
        self.NaOH = Component()
        self.EthylAcetate = Component()
        self.SodiumAcetate = Component()
        self.Ethanol = Component()

        # Heat capacity of water
        self.cp_mol = Param(
            mutable=True,
            initialize=75.327,
            doc="Molar heat capacity of water [J/mol.K]",
            units=units.J / units.mol / units.K,
        )

        # Density of water
        self.dens_mol = Param(
            mutable=True,
            initialize=55388.0,
            doc="Molar density of water [mol/m^3]",
            units=units.mol / units.m**3,
        )

        # Thermodynamic reference state
        self.pressure_ref = Param(
            within=PositiveReals,
            mutable=True,
            default=101325.0,
            doc="Reference pressure [Pa]",
            units=units.Pa,
        )
        self.temperature_ref = Param(
            within=PositiveReals,
            mutable=True,
            default=298.15,
            doc="Reference temperature [K]",
            units=units.K,
        )

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "flow_vol": {"method": None, "units": "m^3/s"},
                "pressure": {"method": None, "units": "Pa"},
                "temperature": {"method": None, "units": "K"},
                "conc_mol_comp": {"method": None, "units": "mol/m^3"},
                "dens_mol": {"method": None, "units": "mol/m^3"},
            }
        )
        obj.add_default_units(
            {
                "time": units.s,
                "length": units.m,
                "mass": units.kg,
                "amount": units.mol,
                "temperature": units.K,
            }
        )


class SaponificationPropertiesScaler(CustomScalerBase):
    """
    Scaler for saponification properties package.

    Flow and concentration are scaled by default value (if no user input provided),
    pressure is scaled assuming order of magnitude of 1e5 Pa, and temperature is
    scaled using the average of the bounds. Constraints using the inverse maximum
    scheme.
    """

    UNIT_SCALING_FACTORS = {
        # "QuantityName: (reference units, scaling factor)
        "Pressure": (units.Pa, 1e-5),
    }

    DEFAULT_SCALING_FACTORS = {
        "flow_vol": 1e2,
        "conc_mol_comp": 1e-2,
    }

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        self.scale_variable_by_default(model.flow_vol, overwrite=overwrite)
        self.scale_variable_by_units(model.pressure, overwrite=overwrite)
        self.scale_variable_by_bounds(model.temperature, overwrite=overwrite)
        for k, v in model.conc_mol_comp.items():
            if k == "H2O":
                self.set_variable_scaling_factor(v, 1e-4, overwrite=overwrite)
            else:
                self.scale_variable_by_default(v, overwrite=overwrite)

    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        if model.is_property_constructed("conc_water_eqn"):
            self.set_constraint_scaling_factor(
                model.conc_water_eqn,
                1e-4,
                overwrite=overwrite,
            )


class _StateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    default_scaler = SaponificationPropertiesScaler

    def fix_initialization_states(self):
        """
        Fixes state variables for state blocks.

        Returns:
            None
        """
        # Fix state variables
        fix_state_vars(self)

        # Constraint on water concentration at outlet - unfix in these cases
        for b in self.values():
            if b.config.defined_state is False:
                b.conc_mol_comp["H2O"].unfix()

    def initialize(
        blk,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for property package.

        Keyword Arguments:
        state_args : Dictionary with initial guesses for the state vars
                     chosen. Note that if this method is triggered
                     through the control volume, and if initial guesses
                     were not provided at the unit model level, the
                     control volume passes the inlet values as initial
                     guess.The keys for the state_args dictionary are:

                     flow_mol_comp : value at which to initialize component
                                     flows (default=None)
                     pressure : value at which to initialize pressure
                                (default=None)
                     temperature : value at which to initialize temperature
                                  (default=None)
            outlvl : sets output level of initialization routine
            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed and
                                       initialization does not need to worry
                                       about fixing and unfixing variables.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states variables are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 release_state method

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="properties")
        # Deactivate the constraints specific for outlet block i.e.
        # when defined state is False
        # This is needed as fixing state vars fixes conc_mol_comp["H2O"],
        # which is also specified by the conc_water_eqn constraint
        for k in blk.values():
            if k.config.defined_state is False:
                k.conc_water_eqn.deactivate()

        if state_vars_fixed is False:
            # Fix state variables if not already fixed
            flags = fix_state_vars(blk, state_args)

        else:
            # Check when the state vars are fixed already result in dof 0
            for k in blk.values():
                if degrees_of_freedom(k) != 0:
                    # PYLINT-TODO
                    # pylint: disable-next=broad-exception-raised
                    raise Exception(
                        "State vars fixed but degrees of freedom "
                        "for state block is not zero during "
                        "initialization."
                    )

        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                blk.release_state(flags)

        init_log.info("Initialization Complete.")

    def release_state(blk, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="properties")

        # Reactivate conc_water_eqn
        for k in blk.values():
            if not k.config.defined_state:
                k.conc_water_eqn.activate()

        if flags is None:
            return
        # Unfix state variables
        revert_state_vars(blk, flags)
        init_log.info("State Released.")


@declare_process_block_class("SaponificationStateBlock", block_class=_StateBlock)
class SaponificationStateBlockData(StateBlockData):
    """
    An example property package for properties for saponification of ethyl
    acetate
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(SaponificationStateBlockData, self).build()

        # Create state variables
        self.flow_vol = Var(
            initialize=1.0,
            domain=NonNegativeReals,
            doc="Total volumentric flowrate [m^3/s]",
            units=units.m**3 / units.s,
        )
        self.pressure = Var(
            domain=Reals,
            initialize=101325.0,
            bounds=(1e3, 1e6),
            doc="State pressure [Pa]",
            units=units.Pa,
        )
        self.temperature = Var(
            domain=Reals,
            initialize=298.15,
            bounds=(298.15, 323.15),
            doc="State temperature [K]",
            units=units.K,
        )
        self.conc_mol_comp = Var(
            self.params.component_list,
            domain=NonNegativeReals,
            initialize=100.0,
            doc="Component molar concentrations " "[mol/m^3]",
            units=units.mol / units.m**3,
        )

        if self.config.defined_state is False:
            self.conc_water_eqn = Constraint(
                expr=self.conc_mol_comp["H2O"] == self.params.dens_mol
            )

    def get_material_flow_terms(b, p, j):
        return b.flow_vol * b.conc_mol_comp[j]

    def get_enthalpy_flow_terms(b, p):
        return (
            b.flow_vol
            * b.params.dens_mol
            * b.params.cp_mol
            * (b.temperature - b.params.temperature_ref)
        )

    def get_material_density_terms(b, p, j):
        return b.conc_mol_comp[j]

    def get_energy_density_terms(b, p):
        return (
            b.params.dens_mol
            * b.params.cp_mol
            * (b.temperature - b.params.temperature_ref)
        )

    def default_material_balance_type(self):
        return MaterialBalanceType.componentPhase

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(b):
        return {
            "flow_vol": b.flow_vol,
            "conc_mol_comp": b.conc_mol_comp,
            "temperature": b.temperature,
            "pressure": b.pressure,
        }

    def define_display_vars(b):
        return {
            "Volumetric Flowrate": b.flow_vol,
            "Molar Concentration": b.conc_mol_comp,
            "Temperature": b.temperature,
            "Pressure": b.pressure,
        }

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def model_check(blk):
        """
        Model checks for property block
        """
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            _log.error("{} Temperature set below lower bound.".format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            _log.error("{} Temperature set above upper bound.".format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            _log.error("{} Pressure set below lower bound.".format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            _log.error("{} Pressure set above upper bound.".format(blk.name))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest

import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentSet
import pyomo.environ as pyo
from pyomo.contrib.incidence_analysis import (
    IncidenceGraphInterface,
)
from pyomo.util.check_units import assert_units_consistent

import idaes.core as idaes
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models_extra.gas_distribution.properties.natural_gas import (
    NaturalGasParameterBlock,
)
from idaes.models_extra.gas_distribution.unit_models.pipeline import GasPipeline
from idaes.models_extra.gas_distribution.unit_models.node import PipelineNode

"""
Tests for the PipelineNode unit model
"""


@pytest.mark.unit
class TestConstructNode(unittest.TestCase):
    """
    Tests for the construction of nodes with various attached pipeline
    configurations.
    """

    def test_construct_node_with_pipelines(self):
        """
        A node with one inlet, one outlet, no supplies, and
        no demands.
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline1 = GasPipeline(**pipeline_config)
        pipeline1 = m.fs.pipeline1
        m.fs.pipeline2 = GasPipeline(**pipeline_config)
        pipeline2 = m.fs.pipeline2

        node_config = {"property_package": m.fs.properties}
        m.fs.node = PipelineNode(**node_config)
        node = m.fs.node

        var_set = ComponentSet(node.component_data_objects(pyo.Var))
        # Three state blocks, four variables each.
        self.assertEqual(len(var_set), 12)

        igraph = IncidenceGraphInterface(node)
        # This test asserts that each of these variables participates
        # in an active constraint (is recognized by the incidence graph)
        self.assertEqual(len(igraph.variables), len(var_set))
        for var in igraph.variables:
            self.assertIn(var, var_set)

        time = m.fs.time
        t0 = time.first()
        j = next(iter(m.fs.properties.component_list))

        states = [node.state, node.inlets[0].state, node.outlets[0].state]
        for state in states:
            # Make sure the inlet and outlet blocks have state blocks
            # with their state variables.
            self.assertTrue(isinstance(state[t0].temperature, pyo.Var))
            self.assertIn(state[t0].temperature, var_set)
            self.assertTrue(isinstance(state[t0].pressure, pyo.Var))
            self.assertIn(state[t0].pressure, var_set)
            self.assertTrue(isinstance(state[t0].mole_frac_comp, pyo.Var))
            self.assertIn(state[t0].mole_frac_comp[j], var_set)
            self.assertTrue(isinstance(state[t0].flow_mol, pyo.Var))
            self.assertIn(state[t0].flow_mol, var_set)

        con_set = ComponentSet(node.component_data_objects(pyo.Constraint))
        # Pressure equality between inlet and node, temperature, pressure,
        # and mole fraction equalities between the outlet and node,
        # plus flow balance, component mixing, and total flow equations.
        self.assertEqual(len(con_set), 8)

        blocks = [node.inlets[0], node.outlets[0]]
        for block in blocks:
            self.assertTrue(isinstance(block.pressure_eq, pyo.Constraint))
            self.assertIn(block.pressure_eq[t0], con_set)

        # Inlet mole fractions and temperatures are already specified,
        # so we don't have this equation.
        block = node.outlets[0]
        self.assertTrue(isinstance(block.mole_frac_comp_eq, pyo.Constraint))
        self.assertIn(block.mole_frac_comp_eq[t0, j], con_set)
        self.assertTrue(isinstance(block.temperature_eq, pyo.Constraint))
        self.assertIn(block.temperature_eq[t0], con_set)

        self.assertTrue(isinstance(node.flow_balance, pyo.Constraint))
        self.assertTrue(isinstance(node.component_mixing_eq, pyo.Constraint))
        self.assertTrue(isinstance(node.total_flow_eq, pyo.Constraint))
        self.assertTrue(isinstance(node.enthalpy_mixing_eq, pyo.Constraint))
        self.assertIn(node.flow_balance[t0], con_set)
        self.assertIn(node.component_mixing_eq[t0, j], con_set)
        self.assertIn(node.total_flow_eq[t0], con_set)
        self.assertIn(node.enthalpy_mixing_eq[t0], con_set)

        node.add_pipeline_to_inlet(pipeline1)
        node.add_pipeline_to_outlet(pipeline2)

        p1 = node.get_inlet_pipeline(0)
        self.assertIs(p1, pipeline1)
        p2 = node.get_outlet_pipeline(0)
        self.assertIs(p2, pipeline2)

        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        pipeline1.diameter.fix()
        pipeline1.control_volume.length.fix()
        pipeline2.diameter.fix()
        pipeline2.control_volume.length.fix()

        cv1 = pipeline1.control_volume
        x0 = cv1.length_domain.first()
        cv1.properties[t0, x0].temperature.fix()
        cv1.properties[t0, x0].pressure.fix()
        cv1.properties[t0, x0].mole_frac_comp[j].fix()
        cv1.properties[t0, x0].flow_mol.fix()

        assert_units_consistent(m)
        igraph = IncidenceGraphInterface(m)
        matching = igraph.maximum_matching()
        N, M = len(igraph.constraints), len(igraph.variables)
        self.assertEqual(degrees_of_freedom(m), 0)
        self.assertEqual(N, M)  # Sanity check
        self.assertEqual(len(matching), N)

    def test_pipeline_with_two_nodes(self):
        """
        One pipeline connecting two nodes.
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)
        pipeline = m.fs.pipeline

        node_configs = [
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 0,
                "n_outlet_pipelines": 1,
                "n_supplies": 1,
                "n_demands": 0,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 1,
                "n_outlet_pipelines": 0,
                "n_supplies": 0,
                "n_demands": 1,
            },
        ]
        m.fs.node_set = pyo.Set(initialize=list(range(len(node_configs))))
        node_configs = {i: config for i, config in enumerate(node_configs)}
        m.fs.nodes = PipelineNode(m.fs.node_set, initialize=node_configs)

        time = m.fs.time
        t0 = m.fs.time.first()

        m.fs.nodes[0].add_pipeline_to_outlet(pipeline)
        m.fs.nodes[1].add_pipeline_to_inlet(pipeline)
        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        # Fix degrees of freedom
        pipeline.control_volume.length.fix()
        pipeline.diameter.fix()

        # Fix state variables at inlet node.
        # Fixing the flow rate here, which represents the total flow rate
        # through the node, implicitly fixes the supply flow.
        # Same goes for the mole fraction.
        #
        # This loop also specifies pressure and temperature of the supply
        for var in m.fs.nodes[0].state[t0].component_data_objects(pyo.Var):
            var.fix()

        assert_units_consistent(m)
        self.assertEqual(degrees_of_freedom(m), 0)
        igraph = IncidenceGraphInterface(m)
        N, M = len(igraph.constraints), len(igraph.variables)
        matching = igraph.maximum_matching()
        self.assertEqual(N, M)
        self.assertEqual(N, len(matching))

        var_set = ComponentSet(igraph.variables)
        self.assertIn(m.fs.nodes[0].supplies[0].flow_mol[t0], var_set)
        self.assertIn(m.fs.nodes[1].demands[0].flow_mol[t0], var_set)

        # Temperature and pressure equality constraints make sure
        # supply variables are the same as node variables.
        self.assertIn(m.fs.nodes[0].supplies[0].state[t0].temperature, var_set)
        self.assertIn(m.fs.nodes[0].supplies[0].state[t0].pressure, var_set)

    def test_multiple_inlets_outlets(self):
        """
        A test with multiple inlets and outlets
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline_set = pyo.Set(initialize=list(range(7)))
        m.fs.pipeline = GasPipeline(m.fs.pipeline_set, **pipeline_config)

        node_config = {
            "property_package": m.fs.properties,
            "n_supplies": 1,
            "n_demands": 2,
            "n_inlet_pipelines": 3,
            "n_outlet_pipelines": 4,
        }
        m.fs.node = PipelineNode(**node_config)

        m.fs.node.add_pipeline_to_inlet(m.fs.pipeline[0])
        m.fs.node.add_pipeline_to_inlet(m.fs.pipeline[1])
        m.fs.node.add_pipeline_to_inlet(m.fs.pipeline[2])
        m.fs.node.add_pipeline_to_outlet(m.fs.pipeline[3])
        m.fs.node.add_pipeline_to_outlet(m.fs.pipeline[4])
        m.fs.node.add_pipeline_to_outlet(m.fs.pipeline[5])
        m.fs.node.add_pipeline_to_outlet(m.fs.pipeline[6])

        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        j = next(iter(m.fs.properties.component_list))

        for pipeline in m.fs.pipeline.values():
            pipeline.control_volume.length.fix()
            pipeline.diameter.fix()

        t0 = m.fs.time.first()
        x0 = m.fs.pipeline[0].control_volume.length_domain.first()
        state0 = m.fs.pipeline[0].control_volume.properties[t0, x0]

        # Fix pressure of one inlet
        state0.pressure.fix()

        # Fix flow rates and mole fractions of all inlets
        for pipeline in m.fs.node.inlet_pipelines().values():
            pipeline.control_volume.flow_mass[t0, x0].fix()
            state = pipeline.control_volume.properties[t0, x0]
            state.mole_frac_comp[j].fix()
            state.temperature.fix()

        # Fix supply flows and mole fractions
        m.fs.node.supplies[:].flow_mol[:].fix()
        m.fs.node.supplies[:].state[:].mole_frac_comp[:].fix()
        m.fs.node.supplies[:].state[:].temperature.fix()

        # Fix demand flows, and all outlet flow rates except one
        m.fs.node.demands[:].flow_mol[:].fix()
        m.fs.pipeline[4].control_volume.flow_mass[t0, x0].fix()
        m.fs.pipeline[5].control_volume.flow_mass[t0, x0].fix()
        m.fs.pipeline[6].control_volume.flow_mass[t0, x0].fix()

        assert_units_consistent(m)
        self.assertEqual(degrees_of_freedom(m), 0)
        igraph = IncidenceGraphInterface(m)
        N, M = len(igraph.constraints), len(igraph.variables)
        matching = igraph.maximum_matching()
        self.assertEqual(N, M)
        self.assertEqual(N, len(matching))

    def test_dynamic_two_nodes(self):
        """
        A dynamic instance of the test with two nodes
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)
        pipeline = m.fs.pipeline

        node_configs = [
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 0,
                "n_outlet_pipelines": 1,
                "n_supplies": 1,
                "n_demands": 0,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 1,
                "n_outlet_pipelines": 0,
                "n_supplies": 0,
                "n_demands": 1,
            },
        ]
        m.fs.node_set = pyo.Set(initialize=list(range(len(node_configs))))
        node_configs = {i: config for i, config in enumerate(node_configs)}
        m.fs.nodes = PipelineNode(m.fs.node_set, initialize=node_configs)

        time = m.fs.time
        t0 = m.fs.time.first()
        ntfe = 4
        disc = pyo.TransformationFactory("dae.finite_difference")
        disc.apply_to(m, wrt=time, nfe=ntfe, scheme="BACKWARD")

        m.fs.nodes[0].add_pipeline_to_outlet(pipeline)
        m.fs.nodes[1].add_pipeline_to_inlet(pipeline)
        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        cv = pipeline.control_volume
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()
        j = next(iter(m.fs.properties.component_list))

        # Deactivate momentum balance where neither derivative is defined
        cv.momentum_balance[t0, xf].deactivate()

        # Fix geometry variables
        pipeline.control_volume.length.fix()
        pipeline.diameter.fix()

        # Fix dynamic inputs: supply pressures and demand flows
        m.fs.nodes[0].supplies[:].state[:].mole_frac_comp[j].fix()
        m.fs.nodes[0].state[:].pressure.fix()
        m.fs.nodes[0].state[:].temperature.fix()
        m.fs.nodes[1].state[:].flow_mol.fix()

        # Fix initial conditions: flow and pressure except where already
        # defined.
        for x in cv.length_domain:
            if x != x0:
                cv.pressure[t0, x].fix()
            if x != xf:
                cv.flow_mass[t0, x].fix()

        self.assertEqual(degrees_of_freedom(m), 0)
        igraph = IncidenceGraphInterface(m)
        N, M = len(igraph.constraints), len(igraph.variables)
        matching = igraph.maximum_matching()
        self.assertEqual(N, M)
        self.assertEqual(N, len(matching))

        var_set = ComponentSet(igraph.variables)
        for t in time:
            self.assertIn(m.fs.nodes[0].supplies[0].flow_mol[t], var_set)
            self.assertIn(m.fs.nodes[0].supplies[0].state[t].temperature, var_set)
            self.assertIn(m.fs.nodes[0].supplies[0].state[t].pressure, var_set)
            self.assertIn(m.fs.nodes[1].demands[0].flow_mol[t], var_set)

    def test_dynamic_multiple_inlet_outlet(self):
        """
        A dynamic instance of the multiple inlet, multiple outlet test
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline_set = pyo.Set(initialize=list(range(7)))
        m.fs.pipeline = GasPipeline(m.fs.pipeline_set, **pipeline_config)

        node_config = {
            "property_package": m.fs.properties,
            "n_supplies": 1,
            "n_demands": 2,
            "n_inlet_pipelines": 3,
            "n_outlet_pipelines": 4,
        }
        m.fs.node = PipelineNode(**node_config)

        m.fs.node.add_pipeline_to_inlet(m.fs.pipeline[0])
        m.fs.node.add_pipeline_to_inlet(m.fs.pipeline[1])
        m.fs.node.add_pipeline_to_inlet(m.fs.pipeline[2])
        m.fs.node.add_pipeline_to_outlet(m.fs.pipeline[3])
        m.fs.node.add_pipeline_to_outlet(m.fs.pipeline[4])
        m.fs.node.add_pipeline_to_outlet(m.fs.pipeline[5])
        m.fs.node.add_pipeline_to_outlet(m.fs.pipeline[6])

        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        time = m.fs.time
        t0 = time.first()
        ntfe = 4
        disc = pyo.TransformationFactory("dae.finite_difference")
        disc.apply_to(m, wrt=time, nfe=ntfe, scheme="BACKWARD")

        x0 = m.fs.pipeline[0].control_volume.length_domain.first()
        xf = m.fs.pipeline[0].control_volume.length_domain.last()
        # Deactivate momentum balance at point where both derivatives
        # are not defined.
        mom_bal = m.fs.pipeline[:].control_volume.momentum_balance[t0, xf]
        mom_bal.deactivate()

        j = next(iter(m.fs.properties.component_list))

        # Fix "static inputs," i.e. design variables
        for pipeline in m.fs.pipeline.values():
            pipeline.control_volume.length.fix()
            pipeline.diameter.fix()

        # Fix dynamic inputs

        state0 = m.fs.pipeline[0].control_volume.properties

        # Pressure of one inlet
        state0[:, x0].pressure.fix()

        # Flow rates, mole fractions, and temperatures of all inlets
        for pipeline in m.fs.node.inlet_pipelines().values():
            pipeline.control_volume.flow_mass[:, x0].fix()
            state = pipeline.control_volume.properties
            state[:, x0].mole_frac_comp[j].fix()
            state[:, x0].temperature.fix()

        # Supply flows, mole fractions, and temperatures
        m.fs.node.supplies[:].flow_mol[:].fix()
        m.fs.node.supplies[:].state[:].mole_frac_comp[:].fix()
        m.fs.node.supplies[:].state[:].temperature.fix()

        # Demand flows, and all outlet flow rates except one
        m.fs.node.demands[:].flow_mol[:].fix()
        m.fs.pipeline[4].control_volume.flow_mass[:, x0].fix()
        m.fs.pipeline[5].control_volume.flow_mass[:, x0].fix()
        m.fs.pipeline[6].control_volume.flow_mass[:, x0].fix()

        # Initial conditions. These are the differential variables of the
        # pipelines, pressure and flow_mass, except where they overlap
        # with "dynamic inputs."
        for pipeline in m.fs.node.inlet_pipelines().values():
            cv = pipeline.control_volume
            state = cv.properties
            for x in cv.length_domain:
                if x != cv.length_domain.first():
                    cv.flow_mass[t0, x].fix()

                if state is state0:
                    # Here, inlet pressure has already been specified
                    if x != cv.length_domain.first():
                        cv.pressure[t0, x].fix()
                else:
                    # Outlet pressure is specified at the initial condition
                    # by node pressure
                    if x != cv.length_domain.last():
                        cv.pressure[t0, x].fix()

        for pipeline in m.fs.node.outlet_pipelines().values():
            cv = pipeline.control_volume
            state = cv.properties
            for x in cv.length_domain:
                if x != cv.length_domain.first():
                    # Only values at the node have already been specified.
                    cv.flow_mass[t0, x].fix()
                    cv.pressure[t0, x].fix()

        igraph = IncidenceGraphInterface(m)
        N, M = len(igraph.constraints), len(igraph.variables)
        matching = igraph.maximum_matching()
        self.assertEqual(degrees_of_freedom(m), 0)
        self.assertEqual(N, M)
        self.assertEqual(N, len(matching))


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest

from idaes.core import FlowsheetBlock
from idaes.models.properties.cubic_eos.cubic_prop_pack import cubic_roots_available
from idaes.models.properties.cubic_eos import BT_PR
from idaes.core.util.exceptions import InitializationError

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Objective,
    value,
)
from pyomo.util.check_units import assert_units_consistent

import idaes.logger as idaeslog
from idaes.models.properties.tests.test_harness import PropertyTestHarness
from idaes.core.solvers import get_solver


# Set module level pyest marker
pytestmark = pytest.mark.cubic_root
prop_available = cubic_roots_available()


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
# Run test harness
@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestBasicLV(PropertyTestHarness):
    def configure(self):
        self.prop_pack = BT_PR.BTParameterBlock
        self.param_args = {"valid_phase": ("Liq", "Vap")}
        self.prop_args = {}
        self.has_density_terms = True


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestBasicL(PropertyTestHarness):
    def configure(self):
        self.prop_pack = BT_PR.BTParameterBlock
        self.param_args = {"valid_phase": "Liq"}
        self.prop_args = {}
        self.has_density_terms = True


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestBasicV(PropertyTestHarness):
    def configure(self):
        self.prop_pack = BT_PR.BTParameterBlock
        self.param_args = {"valid_phase": "Vap"}
        self.prop_args = {}
        self.has_density_terms = True


# -----------------------------------------------------------------------------
# Test robustness and some outputs
@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.skipif(not prop_available, reason="Cubic root finder not available")
class TestBTExample(object):
    @pytest.mark.component
    def test_units(self, caplog):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        with caplog.at_level(idaeslog.WARNING):
            m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))
        assert "May 2025 release." in caplog.text

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        assert_units_consistent(m)

    @pytest.mark.component
    def test_initialization_failure(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(100)
        m.fs.state[0].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[0].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[0].temperature.fix(300)
        m.fs.state[0].pressure.fix(1e5)

        m.fs.state[0].p_bound = Constraint(expr=m.fs.state[0].pressure <= 1e4)

        with pytest.raises(
            InitializationError,
            match="fs.state failed to initialize successfully. "
            "Please check the output logs for more "
            "information.",
        ):
            m.fs.state.initialize()

    @pytest.mark.integration
    def test_T_sweep(self):
        # TODO: This test fails if using MA57 with the linear_presolve
        # This model is known to have some formulation issues.
        solver = get_solver(solver="ipopt_v2", solver_options={"linear_solver": "ma27"})

        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.obj = Objective(expr=(m.fs.state[0].temperature - 510) ** 2)

        for logP in range(8, 13, 1):
            m.fs.obj.deactivate()

            m.fs.state[0].flow_mol.fix(100)
            m.fs.state[0].mole_frac_comp["benzene"].fix(0.5)
            m.fs.state[0].mole_frac_comp["toluene"].fix(0.5)
            m.fs.state[0].temperature.fix(300)
            m.fs.state[0].pressure.fix(10 ** (0.5 * logP))

            m.fs.state.initialize()

            m.fs.state[0].temperature.unfix()
            m.fs.obj.activate()

            results = solver.solve(m, tee=True)

            assert check_optimal_termination(results)
            assert m.fs.state[0].flow_mol_phase["Liq"].value <= 1e-5

    @pytest.mark.integration
    def test_P_sweep(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        for T in range(370, 500, 25):
            m.fs.state[0].flow_mol.fix(100)
            m.fs.state[0].mole_frac_comp["benzene"].fix(0.5)
            m.fs.state[0].mole_frac_comp["toluene"].fix(0.5)
            m.fs.state[0].temperature.fix(T)
            m.fs.state[0].pressure.fix(1e5)

            m.fs.state.initialize()

            results = solver.solve(m)

            assert check_optimal_termination(results)

            while m.fs.state[0].pressure.value <= 1e6:
                m.fs.state[0].pressure.value = m.fs.state[0].pressure.value + 1e5

                results = solver.solve(m)
                assert check_optimal_termination(results)

    @pytest.mark.component
    def test_T350_P1_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(100)
        m.fs.state[0].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[0].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[0].temperature.fix(350)
        m.fs.state[0].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[0].enth_mol_phase
        m.fs.state[0].entr_mol_phase

        m.fs.state.initialize()

        solver.solve(m)

        assert pytest.approx(value(m.fs.state[0]._teq), abs=1e-1) == 365
        assert 0.0035346 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.966749 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.894676
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.347566
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.971072
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.959791
        )

        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.70584
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.29416
        )

        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Liq"]), 1e-5) == 38942.8
        )
        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Vap"]), 1e-5) == 78048.7
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Liq"]), 1e-5) == -367.558
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Vap"]), 1e-5) == -269.0553
        )

    @pytest.mark.component
    def test_T350_P5_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(100)
        m.fs.state[0].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[0].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[0].temperature.fix(350)
        m.fs.state[0].pressure.fix(5e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[0].enth_mol_phase
        m.fs.state[0].entr_mol_phase

        m.fs.state.initialize()

        solver.solve(m)

        assert pytest.approx(value(m.fs.state[0]._teq), 1e-5) == 431.47
        assert (
            pytest.approx(value(m.fs.state[0].compress_fact_phase["Liq"]), 1e-5)
            == 0.01766
        )
        assert (
            pytest.approx(value(m.fs.state[0].compress_fact_phase["Vap"]), 1e-5)
            == 0.80245
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.181229
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.070601
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.856523
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.799237
        )

        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.65415
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.34585
        )

        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Liq"]), 1e-5) == 38966.9
        )
        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Vap"]), 1e-5) == 75150.7
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Liq"]), 1e-5) == -367.6064
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Vap"]), 1e-5) == -287.3318
        )

    @pytest.mark.component
    def test_T450_P1_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(100)
        m.fs.state[0].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[0].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[0].temperature.fix(450)
        m.fs.state[0].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[0].enth_mol_phase
        m.fs.state[0].entr_mol_phase

        m.fs.state.initialize()

        solver.solve(m)

        assert pytest.approx(value(m.fs.state[0]._teq), 1e-5) == 371.4
        assert 0.0033583 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.9821368 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 8.069323
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 4.304955
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.985365
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.979457
        )

        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.29861
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.70139
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.5
        )

        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Liq"]), 1e-5) == 49441.2
        )
        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Vap"]), 1e-5) == 84175.1
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Liq"]), 1e-5) == -333.836
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Vap"]), 1e-5) == -247.385
        )

    @pytest.mark.component
    def test_T450_P5_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(100)
        m.fs.state[0].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[0].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[0].temperature.fix(450)
        m.fs.state[0].pressure.fix(5e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[0].enth_mol_phase
        m.fs.state[0].entr_mol_phase

        m.fs.state.initialize()

        solver.solve(m)

        assert pytest.approx(value(m.fs.state[0]._teq), 1e-5) == 436.93
        assert 0.0166181 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.9053766 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 1.63308
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.873213
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.927534
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.898324
        )

        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.3488737
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.6511263
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.5
        )

        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Liq"]), 1e-5) == 51095.2
        )
        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Vap"]), 1e-5) == 83362.3
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Liq"]), 1e-5) == -331.676
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Vap"]), 1e-5) == -261.961
        )

    @pytest.mark.component
    def test_T368_P1_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(100)
        m.fs.state[0].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[0].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[0].temperature.fix(368)
        m.fs.state[0].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[0].enth_mol_phase
        m.fs.state[0].entr_mol_phase

        m.fs.state.initialize()

        solver.solve(m)

        assert pytest.approx(value(m.fs.state[0]._teq), 1e-5) == 368
        assert 0.003504 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.97 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 1.492049
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.621563
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.97469
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.964642
        )

        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.4012128
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.5987872
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.6141738
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.3858262
        )

        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Liq"]), 1e-5) == 38235.1
        )
        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Vap"]), 1e-5) == 77155.4
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Liq"]), 1e-5) == -364.856
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Vap"]), 1e-5) == -267.892
        )

    @pytest.mark.component
    def test_T376_P1_x2(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = BT_PR.BTParameterBlock(valid_phase=("Vap", "Liq"))

        m.fs.state = m.fs.props.build_state_block([0], defined_state=True)

        m.fs.state[0].flow_mol.fix(100)
        m.fs.state[0].mole_frac_comp["benzene"].fix(0.2)
        m.fs.state[0].mole_frac_comp["toluene"].fix(0.8)
        m.fs.state[0].temperature.fix(376)
        m.fs.state[0].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[0].enth_mol_phase
        m.fs.state[0].entr_mol_phase

        m.fs.state.initialize()

        solver.solve(m)

        assert pytest.approx(value(m.fs.state[0]._teq), 1e-5) == 376
        assert 0.00361333 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.968749 == pytest.approx(
            value(m.fs.state[0].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 1.8394188
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.7871415
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.9763608
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.9663611
        )

        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.17342
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.82658
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.3267155
        )
        assert (
            pytest.approx(
                value(m.fs.state[0].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.6732845
        )

        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Liq"]), 1e-5) == 31535.8
        )
        assert (
            pytest.approx(value(m.fs.state[0].enth_mol_phase["Vap"]), 1e-5) == 69175.3
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Liq"]), 1e-5) == -372.869
        )
        assert (
            pytest.approx(value(m.fs.state[0].entr_mol_phase["Vap"]), 1e-5) == -278.766
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for idaes.commands
"""
# stdlib
from functools import partial
import json
import logging
import os
from pathlib import Path
from shutil import rmtree
import subprocess
import sys
from typing import Union
import uuid

# third-party
from click.testing import CliRunner
import pytest

# package
from idaes.commands import examples, extensions, convergence, config, env_info, base
from . import create_module_scratch, rmtree_scratch
import idaes

__author__ = "Dan Gunter"


_log = logging.getLogger(__name__)
if os.environ.get("IDAES_TEST_DEBUG", False):
    _log.setLevel(logging.DEBUG)


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


scratch_path: Union[Path, None] = None


def setup_module(module):
    global scratch_path
    scratch_path = create_module_scratch(module.__name__)
    _log.info(f"Using scratch dir: {scratch_path}")


def teardown_module(module):
    _log.info(f"Remove files from: {scratch_path}")
    rmtree_scratch(scratch_path)


@pytest.fixture
def tempdir(request):
    function_name = request.function.__name__[5:]  # remove "test_" prefix
    sub_path = scratch_path / function_name
    if sub_path.exists():
        for f in sub_path.glob("*"):
            rmtree(f)
    else:
        sub_path.mkdir()
    return sub_path


class TestBaseCommand:
    @pytest.fixture
    def run_idaes(self, runner):
        return partial(runner.invoke, base.command_base)

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "flags",
        [
            ["--version"],
            ["--help"],
        ],
        ids=" ".join,
    )
    def test_flags(self, run_idaes, flags):
        result = run_idaes(flags)
        assert result.exit_code == 0


################
# get-examples #
################


@pytest.mark.unit
def test_examples_cli(runner):
    result = runner.invoke(examples.get_examples)
    assert result.exit_code == 0


##################
# get-extensions #
##################
@pytest.mark.integration
def test_get_extensions(runner):
    result = runner.invoke(extensions.get_extensions, ["--no-download"])
    assert result.exit_code == 0


@pytest.mark.integration
def test_print_extensions_version(runner):
    result = runner.invoke(extensions.extensions_version)
    assert result.exit_code == 0


@pytest.mark.integration
def test_get_extensions_plat(runner):
    result = runner.invoke(extensions.bin_platform)
    assert result.exit_code == 0


@pytest.mark.integration
def test_get_extensions_bad_plat(runner):
    result = runner.invoke(extensions.bin_platform, ["--distro", "johns_good_linux42"])
    assert result.exit_code == 0
    assert result.output == "No supported binaries found for johns_good_linux42.\n"


@pytest.mark.integration
def test_extensions_license(runner):
    result = runner.invoke(extensions.extensions_license)
    assert result.exit_code == 0


###########
# config  #
###########


@pytest.mark.unit
def test_conf_display(runner):
    result = runner.invoke(config.config_display)
    assert result.exit_code == 0


@pytest.mark.unit
def test_conf_file_paths(runner):
    result = runner.invoke(config.config_file)
    assert result.exit_code == 0


@pytest.mark.unit
def test_conf_file_paths(runner):
    result = runner.invoke(config.config_write)
    assert result.exit_code == 0


@pytest.mark.skipif(sys.version_info[:2] == (3, 11), reason="Fails on Python 3.11")
@pytest.mark.unit
def test_conf_set(runner):
    result = runner.invoke(config.config_set)
    assert result.exit_code == 0


##############
# env info   #
##############


@pytest.mark.unit
def test_env_info1(runner):
    result = runner.invoke(env_info.environment_info)
    assert result.exit_code == 0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for generic property package core code

Author: Andrew Lee
"""
import pytest

# Import Pyomo units
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Expression,
    units as pyunits,
    value,
)

# Import IDAES cores
from idaes.core import (
    FlowsheetBlock,
    VaporPhase,
    Component,
    MaterialBalanceType,
    ControlVolume1DBlock,
)

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
    StateIndex,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.pure import RPP4
from idaes.models.properties.modular_properties.reactions.dh_rxn import constant_dh_rxn
from idaes.models.properties.modular_properties.reactions.equilibrium_constant import (
    van_t_hoff,
)
from idaes.models.properties.modular_properties.reactions.equilibrium_forms import (
    power_law_equil,
)
from idaes.models.properties.modular_properties.base.utility import ConcentrationForm

from idaes.models.unit_models import Heater
from idaes.core.util.model_statistics import degrees_of_freedom

from idaes.core.solvers import get_solver

solver = get_solver("ipopt_v2")


configuration = {
    # Specifying components
    "components": {
        "a": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "parameter_data": {
                "mw": (78.1136e-3, pyunits.kg / pyunits.mol),
                "pressure_crit": (48.9e5, pyunits.Pa),
                "temperature_crit": (562.2, pyunits.K),
                "cp_mol_ig_comp_coeff": {
                    "A": (-40, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (0, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (0, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (0, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol),
            },
        },
        "b": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "parameter_data": {
                "mw": (78.1136e-3, pyunits.kg / pyunits.mol),
                "pressure_crit": (48.9e5, pyunits.Pa),
                "temperature_crit": (562.2, pyunits.K),
                "cp_mol_ig_comp_coeff": {
                    "A": (-40, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (0, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (0, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (0, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol),
            },
        },
        "c": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "parameter_data": {
                "mw": (78.1136e-3, pyunits.kg / pyunits.mol),
                "pressure_crit": (48.9e5, pyunits.Pa),
                "temperature_crit": (562.2, pyunits.K),
                "cp_mol_ig_comp_coeff": {
                    "A": (-40, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (0, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (0, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (0, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol),
            },
        },
        "d": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "parameter_data": {
                "mw": (78.1136e-3, pyunits.kg / pyunits.mol),
                "pressure_crit": (48.9e5, pyunits.Pa),
                "temperature_crit": (562.2, pyunits.K),
                "cp_mol_ig_comp_coeff": {
                    "A": (-40, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (0, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (0, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (0, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol),
            },
        },
    },
    # Specifying phases
    "phases": {"Vap": {"type": VaporPhase, "equation_of_state": Ideal}},
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 450, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
    },
    "state_components": StateIndex.true,  # need this for inherent rxn terms in material balances
    "pressure_ref": (1e5, pyunits.Pa),
    "temperature_ref": (300, pyunits.K),
    "inherent_reactions": {
        "e1": {
            "stoichiometry": {("Vap", "a"): -1, ("Vap", "b"): 1},
            "heat_of_reaction": constant_dh_rxn,
            "equilibrium_constant": van_t_hoff,
            "equilibrium_form": power_law_equil,
            "concentration_form": ConcentrationForm.moleFraction,
            "parameter_data": {"dh_rxn_ref": -20000, "k_eq_ref": 2, "T_eq_ref": 350},
        }
    },
}


class TestInherentReactions(object):
    @pytest.fixture()
    def frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = GenericParameterBlock(**configuration)

        return m

    @pytest.mark.unit
    def test_build(self, frame):
        frame.fs.props = frame.fs.params.build_state_block([1], defined_state=False)

        assert isinstance(frame.fs.props[1].inherent_equilibrium_constraint, Constraint)
        assert len(frame.fs.props[1].inherent_equilibrium_constraint) == 1

        assert isinstance(frame.fs.props[1].dh_rxn, Expression)
        assert len(frame.fs.props[1].dh_rxn) == 1

    @pytest.mark.component
    def test_heater_w_inherent_rxns_comp_phase(self, frame):
        frame.fs.H101 = Heater(
            property_package=frame.fs.params,
            material_balance_type=MaterialBalanceType.componentPhase,
        )

        frame.fs.H101.inlet.flow_mol.fix(100)
        frame.fs.H101.inlet.mole_frac_comp.fix(0.25)
        frame.fs.H101.inlet.temperature.fix(350)
        frame.fs.H101.inlet.pressure.fix(101325)

        frame.fs.H101.heat_duty.fix(0)

        assert (degrees_of_freedom(frame)) == 0

        frame.fs.H101.initialize()

        results = solver.solve(frame)

        assert check_optimal_termination(results)

        assert value(
            frame.fs.H101.control_volume.properties_out[0].k_eq["e1"]
        ) == pytest.approx(2, rel=1e-8)

        assert value(
            frame.fs.H101.control_volume.properties_out[0].k_eq["e1"]
        ) == pytest.approx(
            value(
                frame.fs.H101.outlet.mole_frac_comp[0, "b"]
                / frame.fs.H101.outlet.mole_frac_comp[0, "a"]
            ),
            rel=1e-5,
        )

        assert value(frame.fs.H101.outlet.mole_frac_comp[0, "a"]) == pytest.approx(
            1 / 6, rel=1e-5
        )
        assert value(frame.fs.H101.outlet.mole_frac_comp[0, "b"]) == pytest.approx(
            1 / 3, rel=1e-5
        )
        assert value(frame.fs.H101.outlet.mole_frac_comp[0, "c"]) == pytest.approx(
            1 / 4, rel=1e-5
        )
        assert value(frame.fs.H101.outlet.mole_frac_comp[0, "d"]) == pytest.approx(
            1 / 4, rel=1e-5
        )

        assert value(frame.fs.H101.outlet.flow_mol[0]) == pytest.approx(100, rel=1e-5)
        assert value(frame.fs.H101.outlet.temperature[0]) == pytest.approx(
            350, rel=1e-5
        )
        assert value(frame.fs.H101.outlet.pressure[0]) == pytest.approx(
            101325, rel=1e-5
        )

    @pytest.mark.component
    def test_heater_w_inherent_rxns_comp_total(self, frame):
        frame.fs.H101 = Heater(
            property_package=frame.fs.params,
            material_balance_type=MaterialBalanceType.componentTotal,
        )

        frame.fs.H101.inlet.flow_mol.fix(100)
        frame.fs.H101.inlet.mole_frac_comp.fix(0.25)
        frame.fs.H101.inlet.temperature.fix(350)
        frame.fs.H101.inlet.pressure.fix(101325)

        frame.fs.H101.heat_duty.fix(0)

        assert (degrees_of_freedom(frame)) == 0

        frame.fs.H101.initialize()

        results = solver.solve(frame)

        assert check_optimal_termination(results)

        assert value(
            frame.fs.H101.control_volume.properties_out[0].k_eq["e1"]
        ) == pytest.approx(2, rel=1e-8)

        assert value(
            frame.fs.H101.control_volume.properties_out[0].k_eq["e1"]
        ) == pytest.approx(
            value(
                frame.fs.H101.outlet.mole_frac_comp[0, "b"]
                / frame.fs.H101.outlet.mole_frac_comp[0, "a"]
            ),
            rel=1e-5,
        )

        assert value(frame.fs.H101.outlet.mole_frac_comp[0, "a"]) == pytest.approx(
            1 / 6, rel=1e-5
        )
        assert value(frame.fs.H101.outlet.mole_frac_comp[0, "b"]) == pytest.approx(
            1 / 3, rel=1e-5
        )
        assert value(frame.fs.H101.outlet.mole_frac_comp[0, "c"]) == pytest.approx(
            1 / 4, rel=1e-5
        )
        assert value(frame.fs.H101.outlet.mole_frac_comp[0, "d"]) == pytest.approx(
            1 / 4, rel=1e-5
        )

        assert value(frame.fs.H101.outlet.flow_mol[0]) == pytest.approx(100, rel=1e-5)
        assert value(frame.fs.H101.outlet.temperature[0]) == pytest.approx(
            350, rel=1e-5
        )
        assert value(frame.fs.H101.outlet.pressure[0]) == pytest.approx(
            101325, rel=1e-5
        )

    @pytest.mark.component
    def test_CV1D_w_inherent_rxns_comp_phase(self, frame):
        frame.fs.cv = ControlVolume1DBlock(
            property_package=frame.fs.params,
            transformation_method="dae.finite_difference",
            transformation_scheme="BACKWARD",
            finite_elements=2,
        )

        frame.fs.cv.add_geometry()

        frame.fs.cv.add_state_blocks(has_phase_equilibrium=False)

        frame.fs.cv.add_material_balances(
            balance_type=MaterialBalanceType.componentPhase
        )

        frame.fs.cv.add_energy_balances()
        frame.fs.cv.add_momentum_balances()
        frame.fs.cv.apply_transformation()

        frame.fs.cv.properties[0, 0].flow_mol.fix(100)
        frame.fs.cv.properties[0, 0].mole_frac_comp.fix(0.25)
        frame.fs.cv.properties[0, 0].temperature.fix(350)
        frame.fs.cv.properties[0, 0].pressure.fix(101325)

        frame.fs.cv.area.fix(1)
        frame.fs.cv.length.fix(1)

        assert (degrees_of_freedom(frame)) == 0

        frame.fs.cv.initialize()

        results = solver.solve(frame)

        assert check_optimal_termination(results)

        assert pytest.approx(2, rel=1e-8) == value(
            frame.fs.cv.properties[0, 1].k_eq["e1"]
        )

        assert value(frame.fs.cv.properties[0, 1].k_eq["e1"]) == pytest.approx(
            value(
                frame.fs.cv.properties[0, 1].mole_frac_comp["b"]
                / frame.fs.cv.properties[0, 1].mole_frac_comp["a"]
            ),
            rel=1e-5,
        )

        assert value(frame.fs.cv.properties[0, 1].mole_frac_comp["a"]) == pytest.approx(
            1 / 6, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].mole_frac_comp["b"]) == pytest.approx(
            1 / 3, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].mole_frac_comp["c"]) == pytest.approx(
            1 / 4, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].mole_frac_comp["d"]) == pytest.approx(
            1 / 4, rel=1e-5
        )

        assert value(frame.fs.cv.properties[0, 1].flow_mol) == pytest.approx(
            100, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].temperature) == pytest.approx(
            350, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].pressure) == pytest.approx(
            101325, rel=1e-5
        )

    @pytest.mark.component
    def test_CV1D_w_inherent_rxns_comp_total(self, frame):
        frame.fs.cv = ControlVolume1DBlock(
            property_package=frame.fs.params,
            transformation_method="dae.finite_difference",
            transformation_scheme="BACKWARD",
            finite_elements=2,
        )

        frame.fs.cv.add_geometry()

        frame.fs.cv.add_state_blocks(has_phase_equilibrium=False)

        frame.fs.cv.add_material_balances(
            balance_type=MaterialBalanceType.componentTotal
        )

        frame.fs.cv.add_energy_balances()
        frame.fs.cv.add_momentum_balances()
        frame.fs.cv.apply_transformation()

        frame.fs.cv.properties[0, 0].flow_mol.fix(100)
        frame.fs.cv.properties[0, 0].mole_frac_comp.fix(0.25)
        frame.fs.cv.properties[0, 0].temperature.fix(350)
        frame.fs.cv.properties[0, 0].pressure.fix(101325)

        frame.fs.cv.area.fix(1)
        frame.fs.cv.length.fix(1)

        assert (degrees_of_freedom(frame)) == 0

        frame.fs.cv.initialize()

        results = solver.solve(frame)

        assert check_optimal_termination(results)

        assert pytest.approx(2, rel=1e-8) == value(
            frame.fs.cv.properties[0, 1].k_eq["e1"]
        )

        assert value(frame.fs.cv.properties[0, 1].k_eq["e1"]) == pytest.approx(
            value(
                frame.fs.cv.properties[0, 1].mole_frac_comp["b"]
                / frame.fs.cv.properties[0, 1].mole_frac_comp["a"]
            ),
            rel=1e-5,
        )

        assert value(frame.fs.cv.properties[0, 1].mole_frac_comp["a"]) == pytest.approx(
            1 / 6, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].mole_frac_comp["b"]) == pytest.approx(
            1 / 3, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].mole_frac_comp["c"]) == pytest.approx(
            1 / 4, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].mole_frac_comp["d"]) == pytest.approx(
            1 / 4, rel=1e-5
        )

        assert value(frame.fs.cv.properties[0, 1].flow_mol) == pytest.approx(
            100, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].temperature) == pytest.approx(
            350, rel=1e-5
        )
        assert value(frame.fs.cv.properties[0, 1].pressure) == pytest.approx(
            101325, rel=1e-5
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Flash unit model.
Author: Jaffer Ghouse
"""
import pytest

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    value,
    units,
    units as pyunits,
)

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from idaes.models.unit_models.flash import Flash, EnergySplittingType
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
from idaes.models.properties import iapws95
from idaes.core.util.model_statistics import (
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import PhysicalParameterTestBlock, initialization_tester
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    SingleControlVolumeUnitInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = Flash(property_package=m.fs.properties)

    # Check unit config arguments
    assert len(m.fs.unit.config) == 11

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.ideal_separation
    assert m.fs.unit.config.has_heat_transfer
    assert m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.energy_split_basis == EnergySplittingType.equal_temperature


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_calc_scale():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.unit = Flash(property_package=m.fs.properties)
    iscale.calculate_scaling_factors(m)


# -----------------------------------------------------------------------------
class TestBTXIdeal(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = BTXParameterBlock(
            valid_phase=("Liq", "Vap"), activity_coeff_model="Ideal"
        )

        m.fs.unit = Flash(property_package=m.fs.properties)

        m.fs.unit.inlet.flow_mol.fix(1)
        m.fs.unit.inlet.temperature.fix(368)
        m.fs.unit.inlet.pressure.fix(101325)
        m.fs.unit.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        # Legacy property package, does not bound many variables which triggers
        # warnings for potential evaluation errors.
        # Fixing property package is out of scope for now.
        m.fs.unit.control_volume.properties_in[0.0].temperature_bubble.setlb(300)
        m.fs.unit.control_volume.properties_in[0.0].temperature_bubble.setub(550)
        m.fs.unit.control_volume.properties_in[0.0].temperature_dew.setlb(300)
        m.fs.unit.control_volume.properties_in[0.0].temperature_dew.setub(550)
        m.fs.unit.control_volume.properties_in[0.0]._temperature_equilibrium.setlb(300)
        m.fs.unit.control_volume.properties_in[0.0]._temperature_equilibrium.setub(550)
        m.fs.unit.control_volume.properties_in[0.0].pressure_sat_comp.setlb(1e4)
        m.fs.unit.control_volume.properties_in[0.0].pressure_sat_comp.setub(5e6)

        m.fs.unit.control_volume.properties_out[0.0].temperature_bubble.setlb(300)
        m.fs.unit.control_volume.properties_out[0.0].temperature_bubble.setub(550)
        m.fs.unit.control_volume.properties_out[0.0].temperature_dew.setlb(300)
        m.fs.unit.control_volume.properties_out[0.0].temperature_dew.setub(550)
        m.fs.unit.control_volume.properties_out[0.0]._temperature_equilibrium.setlb(300)
        m.fs.unit.control_volume.properties_out[0.0]._temperature_equilibrium.setub(550)
        m.fs.unit.control_volume.properties_out[0.0].pressure_sat_comp.setlb(1e4)
        m.fs.unit.control_volume.properties_out[0.0].pressure_sat_comp.setub(5e6)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, btx):
        assert hasattr(btx.fs.unit.inlet, "flow_mol")
        assert hasattr(btx.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet, "temperature")
        assert hasattr(btx.fs.unit.inlet, "pressure")

        assert hasattr(btx.fs.unit, "vap_outlet")
        assert len(btx.fs.unit.vap_outlet.vars) == 4
        assert hasattr(btx.fs.unit.vap_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.vap_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.vap_outlet, "temperature")
        assert hasattr(btx.fs.unit.vap_outlet, "pressure")

        assert hasattr(btx.fs.unit, "liq_outlet")
        assert len(btx.fs.unit.liq_outlet.vars) == 4
        assert hasattr(btx.fs.unit.liq_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.liq_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.liq_outlet, "temperature")
        assert hasattr(btx.fs.unit.liq_outlet, "pressure")

        assert hasattr(btx.fs.unit, "split")
        assert hasattr(btx.fs.unit, "heat_duty")
        assert hasattr(btx.fs.unit, "deltaP")

        assert number_variables(btx) == 48
        assert number_total_constraints(btx) == 41
        assert number_unused_variables(btx) == 0

    @pytest.mark.component
    def test_structural_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        dt.assert_no_structural_warnings()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, btx):
        perf_dict = btx.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Heat Duty": btx.fs.unit.heat_duty[0],
                "Pressure Change": btx.fs.unit.deltaP[0],
            }
        }

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, btx):
        stable = btx.fs.unit._get_stream_table_contents()

        expected = {
            "Units": {
                "flow_mol": getattr(pyunits.pint_registry, "mole/s"),
                "mole_frac_comp benzene": getattr(
                    pyunits.pint_registry, "dimensionless"
                ),
                "mole_frac_comp toluene": getattr(
                    pyunits.pint_registry, "dimensionless"
                ),
                "temperature": getattr(pyunits.pint_registry, "K"),
                "pressure": getattr(pyunits.pint_registry, "Pa"),
            },
            "Inlet": {
                "flow_mol": pytest.approx(1.00, rel=1e-4),
                "mole_frac_comp benzene": pytest.approx(0.5, rel=1e-4),
                "mole_frac_comp toluene": pytest.approx(0.5, rel=1e-4),
                "temperature": pytest.approx(368, rel=1e-4),
                "pressure": pytest.approx(101325, rel=1e-4),
            },
            "Vapor Outlet": {
                "flow_mol": pytest.approx(0.5, rel=1e-4),
                "mole_frac_comp benzene": pytest.approx(0.5, rel=1e-4),
                "mole_frac_comp toluene": pytest.approx(0.5, rel=1e-4),
                "temperature": pytest.approx(298.15, rel=1e-4),
                "pressure": pytest.approx(101325.0, rel=1e-4),
            },
            "Liquid Outlet": {
                "flow_mol": pytest.approx(0.5, rel=1e-4),
                "mole_frac_comp benzene": pytest.approx(0.5, rel=1e-4),
                "mole_frac_comp toluene": pytest.approx(0.5, rel=1e-4),
                "temperature": pytest.approx(298.15, rel=1e-4),
                "pressure": pytest.approx(101325.0, rel=1e-4),
            },
        }

        assert stable.to_dict() == expected

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx):
        initialization_tester(btx)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx):
        assert pytest.approx(0.603, abs=1e-3) == value(
            btx.fs.unit.liq_outlet.flow_mol[0]
        )
        assert pytest.approx(0.396, abs=1e-3) == value(
            btx.fs.unit.vap_outlet.flow_mol[0]
        )
        assert pytest.approx(368, abs=1e-3) == value(
            btx.fs.unit.liq_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx.fs.unit.liq_outlet.pressure[0]
        )

        assert pytest.approx(0.412, abs=1e-3) == value(
            btx.fs.unit.liq_outlet.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.588, abs=1e-3) == value(
            btx.fs.unit.liq_outlet.mole_frac_comp[0, "toluene"]
        )
        assert pytest.approx(0.634, abs=1e-3) == value(
            btx.fs.unit.vap_outlet.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.366, abs=1e-3) == value(
            btx.fs.unit.vap_outlet.mole_frac_comp[0, "toluene"]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, btx):
        assert (
            abs(
                value(
                    btx.fs.unit.inlet.flow_mol[0]
                    - (
                        btx.fs.unit.vap_outlet.flow_mol[0]
                        + btx.fs.unit.liq_outlet.flow_mol[0]
                    )
                )
            )
            <= 1e-6
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        dt.assert_no_numerical_warnings()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
class TestIAPWS(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = iapws95.Iapws95ParameterBlock(
            phase_presentation=iapws95.PhaseType.LG
        )

        m.fs.unit = Flash(
            property_package=m.fs.properties,
            ideal_separation=False,
            energy_split_basis=EnergySplittingType.enthalpy_split,
        )

        m.fs.unit.inlet.flow_mol.fix(100)
        m.fs.unit.inlet.enth_mol.fix(24000)
        m.fs.unit.inlet.pressure.fix(101325)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iapws):
        assert len(iapws.fs.unit.inlet.vars) == 3
        assert hasattr(iapws.fs.unit.inlet, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet, "pressure")

        assert hasattr(iapws.fs.unit, "vap_outlet")
        assert len(iapws.fs.unit.vap_outlet.vars) == 3
        assert hasattr(iapws.fs.unit.vap_outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.vap_outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.vap_outlet, "pressure")

        assert hasattr(iapws.fs.unit, "liq_outlet")
        assert len(iapws.fs.unit.liq_outlet.vars) == 3
        assert hasattr(iapws.fs.unit.liq_outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.liq_outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.liq_outlet, "pressure")

        assert hasattr(iapws.fs.unit, "split")
        assert hasattr(iapws.fs.unit, "heat_duty")
        assert hasattr(iapws.fs.unit, "deltaP")

        assert number_variables(iapws) == 18
        assert number_total_constraints(iapws) == 13
        assert number_unused_variables(iapws) == 0

    @pytest.mark.component
    def test_structural_issues(self, iapws):
        dt = DiagnosticsToolbox(iapws)
        dt.assert_no_structural_warnings()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, iapws):
        perf_dict = iapws.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Heat Duty": iapws.fs.unit.heat_duty[0],
                "Pressure Change": iapws.fs.unit.deltaP[0],
            }
        }

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, iapws):
        stable = iapws.fs.unit._get_stream_table_contents()

        expected = {
            "Units": {
                "Molar Flow": getattr(pyunits.pint_registry, "mole/second"),
                "Mass Flow": getattr(pyunits.pint_registry, "kg/second"),
                "T": getattr(pyunits.pint_registry, "K"),
                "P": getattr(pyunits.pint_registry, "Pa"),
                "Vapor Fraction": getattr(pyunits.pint_registry, "dimensionless"),
                "Molar Enthalpy": getattr(pyunits.pint_registry, "J/mole"),
            },
            "Inlet": {
                "Molar Flow": pytest.approx(100, rel=1e-4),
                "Mass Flow": pytest.approx(1.8015, rel=1e-4),
                "T": pytest.approx(373.12429584768876, rel=1e-4),
                "P": pytest.approx(101325.0, rel=1e-4),
                "Vapor Fraction": pytest.approx(0.40467852502291135, abs=1e-4),
                "Molar Enthalpy": pytest.approx(24000.0, rel=1e-4),
            },
            "Vapor Outlet": {
                "Molar Flow": pytest.approx(1, rel=1e-4),
                "Mass Flow": pytest.approx(1.8015e-2, rel=1e-4),
                "T": pytest.approx(270.4877112932641, rel=1e-4),
                "P": pytest.approx(11032305.8275, rel=1e-4),
                "Vapor Fraction": pytest.approx(0, abs=1e-4),
                "Molar Enthalpy": pytest.approx(0.01102138712926277, rel=1e-4),
            },
            "Liquid Outlet": {
                "Molar Flow": pytest.approx(1, rel=1e-4),
                "Mass Flow": pytest.approx(1.8015e-2, rel=1e-4),
                "T": pytest.approx(270.4877112932641, rel=1e-4),
                "P": pytest.approx(11032305.8275, rel=1e-4),
                "Vapor Fraction": pytest.approx(0, abs=1e-4),
                "Molar Enthalpy": pytest.approx(0.01102138712926277, rel=1e-4),
            },
        }

        assert stable.to_dict() == expected

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iapws):
        initialization_tester(iapws)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iapws):
        results = solver.solve(iapws)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iapws):
        t_in = value(iapws.fs.unit.control_volume.properties_in[0].temperature)
        assert pytest.approx(373.12, abs=1e-2) == t_in
        assert pytest.approx(t_in, abs=1e-2) == value(
            iapws.fs.unit.control_volume.properties_out[0].temperature
        )
        assert pytest.approx(t_in, abs=1e-2) == value(
            iapws.fs.unit.split.Vap_state[0].temperature
        )
        assert pytest.approx(t_in, abs=1e-2) == value(
            iapws.fs.unit.split.Liq_state[0].temperature
        )

        assert pytest.approx(101325.0, abs=1e3) == value(
            iapws.fs.unit.vap_outlet.pressure[0]
        )
        assert pytest.approx(48200, abs=1e3) == value(
            iapws.fs.unit.vap_outlet.enth_mol[0]
        )
        assert pytest.approx(40.467, abs=1e-3) == value(
            iapws.fs.unit.vap_outlet.flow_mol[0]
        )

        assert pytest.approx(101325.0, abs=1e3) == value(
            iapws.fs.unit.liq_outlet.pressure[0]
        )
        assert pytest.approx(7549.4, abs=1) == value(
            iapws.fs.unit.liq_outlet.enth_mol[0]
        )
        assert pytest.approx(59.532, abs=1e-3) == value(
            iapws.fs.unit.liq_outlet.flow_mol[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iapws):
        assert (
            abs(
                value(
                    iapws.fs.unit.inlet.flow_mol[0]
                    - (
                        iapws.fs.unit.vap_outlet.flow_mol[0]
                        + iapws.fs.unit.liq_outlet.flow_mol[0]
                    )
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    iapws.fs.unit.inlet.flow_mol[0] * iapws.fs.unit.inlet.enth_mol[0]
                    - (
                        iapws.fs.unit.vap_outlet.flow_mol[0]
                        * iapws.fs.unit.vap_outlet.enth_mol[0]
                        + iapws.fs.unit.liq_outlet.flow_mol[0]
                        * iapws.fs.unit.liq_outlet.enth_mol[0]
                    )
                )
            )
            <= 1e-6
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, iapws):
        dt = DiagnosticsToolbox(iapws)
        dt.assert_no_numerical_warnings()


class TestInitializersIAPWS:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = iapws95.Iapws95ParameterBlock(
            phase_presentation=iapws95.PhaseType.LG
        )

        m.fs.unit = Flash(
            property_package=m.fs.properties,
            ideal_separation=False,
            energy_split_basis=EnergySplittingType.enthalpy_split,
        )

        m.fs.unit.inlet.flow_mol.fix(100)
        m.fs.unit.inlet.enth_mol.fix(24000)
        m.fs.unit.inlet.pressure.fix(101325)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.integration
    def test_general_hierarchical(self, model):
        initializer = SingleControlVolumeUnitInitializer()
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        t_in = value(model.fs.unit.control_volume.properties_in[0].temperature)
        assert pytest.approx(373.12, abs=1e-2) == t_in
        assert pytest.approx(t_in, abs=1e-2) == value(
            model.fs.unit.control_volume.properties_out[0].temperature
        )
        assert pytest.approx(t_in, abs=1e-2) == value(
            model.fs.unit.split.Vap_state[0].temperature
        )
        assert pytest.approx(t_in, abs=1e-2) == value(
            model.fs.unit.split.Liq_state[0].temperature
        )

        assert pytest.approx(101325.0, abs=1e3) == value(
            model.fs.unit.vap_outlet.pressure[0]
        )
        assert pytest.approx(48200, abs=1e3) == value(
            model.fs.unit.vap_outlet.enth_mol[0]
        )
        assert pytest.approx(40.467, abs=1e-3) == value(
            model.fs.unit.vap_outlet.flow_mol[0]
        )

        assert pytest.approx(101325.0, abs=1e3) == value(
            model.fs.unit.liq_outlet.pressure[0]
        )
        assert pytest.approx(7549.4, abs=1) == value(
            model.fs.unit.liq_outlet.enth_mol[0]
        )
        assert pytest.approx(59.532, abs=1e-3) == value(
            model.fs.unit.liq_outlet.flow_mol[0]
        )

    @pytest.mark.integration
    def test_block_triangularization(self, model):
        initializer = BlockTriangularizationInitializer(constraint_tolerance=2e-5)
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        t_in = value(model.fs.unit.control_volume.properties_in[0].temperature)
        assert pytest.approx(373.12, abs=1e-2) == t_in
        assert pytest.approx(t_in, abs=1e-2) == value(
            model.fs.unit.control_volume.properties_out[0].temperature
        )
        assert pytest.approx(t_in, abs=1e-2) == value(
            model.fs.unit.split.Vap_state[0].temperature
        )
        assert pytest.approx(t_in, abs=1e-2) == value(
            model.fs.unit.split.Liq_state[0].temperature
        )

        assert pytest.approx(101325.0, abs=1e3) == value(
            model.fs.unit.vap_outlet.pressure[0]
        )
        assert pytest.approx(48200, abs=1e3) == value(
            model.fs.unit.vap_outlet.enth_mol[0]
        )
        assert pytest.approx(40.467, abs=1e-3) == value(
            model.fs.unit.vap_outlet.flow_mol[0]
        )

        assert pytest.approx(101325.0, abs=1e3) == value(
            model.fs.unit.liq_outlet.pressure[0]
        )
        assert pytest.approx(7549.4, abs=1) == value(
            model.fs.unit.liq_outlet.enth_mol[0]
        )
        assert pytest.approx(59.532, abs=1e-3) == value(
            model.fs.unit.liq_outlet.flow_mol[0]
        )


class TestInitializersCubicBTX:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = BTXParameterBlock(
            valid_phase=("Liq", "Vap"), activity_coeff_model="Ideal"
        )

        m.fs.unit = Flash(property_package=m.fs.properties)

        m.fs.unit.inlet.flow_mol[0].set_value(1)
        m.fs.unit.inlet.temperature[0].set_value(368)
        m.fs.unit.inlet.pressure[0].set_value(101325)
        m.fs.unit.inlet.mole_frac_comp[0, "benzene"].set_value(0.5)
        m.fs.unit.inlet.mole_frac_comp[0, "toluene"].set_value(0.5)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.component
    def test_general_hierarchical(self, model):
        initializer = SingleControlVolumeUnitInitializer()
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(0.603, abs=1e-3) == value(
            model.fs.unit.liq_outlet.flow_mol[0]
        )
        assert pytest.approx(0.396, abs=1e-3) == value(
            model.fs.unit.vap_outlet.flow_mol[0]
        )
        assert pytest.approx(368, abs=1e-3) == value(
            model.fs.unit.liq_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            model.fs.unit.liq_outlet.pressure[0]
        )

        assert pytest.approx(0.412, abs=1e-3) == value(
            model.fs.unit.liq_outlet.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.588, abs=1e-3) == value(
            model.fs.unit.liq_outlet.mole_frac_comp[0, "toluene"]
        )
        assert pytest.approx(0.634, abs=1e-3) == value(
            model.fs.unit.vap_outlet.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.366, abs=1e-3) == value(
            model.fs.unit.vap_outlet.mole_frac_comp[0, "toluene"]
        )

        assert not model.fs.unit.inlet.flow_mol[0].fixed
        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "benzene"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "toluene"].fixed

    @pytest.mark.component
    def test_block_triangularization(self, model):
        initializer = BlockTriangularizationInitializer(
            constraint_tolerance=2e-5,
            block_solver_writer_config={"linear_presolve": False},
        )
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(0.603, abs=1e-3) == value(
            model.fs.unit.liq_outlet.flow_mol[0]
        )
        assert pytest.approx(0.396, abs=1e-3) == value(
            model.fs.unit.vap_outlet.flow_mol[0]
        )
        assert pytest.approx(368, abs=1e-3) == value(
            model.fs.unit.liq_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            model.fs.unit.liq_outlet.pressure[0]
        )

        assert pytest.approx(0.412, abs=1e-3) == value(
            model.fs.unit.liq_outlet.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.588, abs=1e-3) == value(
            model.fs.unit.liq_outlet.mole_frac_comp[0, "toluene"]
        )
        assert pytest.approx(0.634, abs=1e-3) == value(
            model.fs.unit.vap_outlet.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.366, abs=1e-3) == value(
            model.fs.unit.vap_outlet.mole_frac_comp[0, "toluene"]
        )

        assert not model.fs.unit.inlet.flow_mol[0].fixed
        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "benzene"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "toluene"].fixed



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest

import pyomo.environ as pyo
from pyomo.common import unittest as pyo_unittest

from idaes.apps.grid_integration.bidder import Bidder
from idaes.apps.grid_integration.tracker import Tracker
from idaes.apps.grid_integration.coordinator import DoubleLoopCoordinator
from idaes.apps.grid_integration.tests.util import (
    ExampleModel,
    ExampleForecaster,
    testing_model_data,
    testing_renewable_data,
    renewable_generator_params,
    testing_generator_params,
)

tracking_horizon = 4
day_ahead_bidding_horizon = 48
real_time_bidding_horizon = tracking_horizon
n_scenario = 3
n_tracking_hour = 1


@pytest.fixture(params=["thermal", "renewable"])
def coordinator_object(request):

    # create solver
    solver = pyo.SolverFactory("cbc")

    ## create trackers
    # make a tracker
    if request.param == "thermal":
        model_data = testing_model_data
    else:
        model_data = testing_renewable_data

    tracking_model_object = ExampleModel(model_data=model_data)
    thermal_tracker = Tracker(
        tracking_model_object=tracking_model_object,
        tracking_horizon=tracking_horizon,
        n_tracking_hour=n_tracking_hour,
        solver=solver,
    )

    # make a projection tracker
    projection_tracking_model_object = ExampleModel(model_data=model_data)
    thermal_projection_tracker = Tracker(
        tracking_model_object=projection_tracking_model_object,
        tracking_horizon=tracking_horizon,
        n_tracking_hour=n_tracking_hour,
        solver=solver,
    )

    ## create a bidder
    forecaster = ExampleForecaster(prediction=30)
    bidding_model_object = ExampleModel(model_data=model_data)
    thermal_bidder = Bidder(
        bidding_model_object=bidding_model_object,
        day_ahead_horizon=day_ahead_bidding_horizon,
        real_time_horizon=real_time_bidding_horizon,
        n_scenario=n_scenario,
        solver=solver,
        forecaster=forecaster,
    )

    ## create coordinator
    coordinator_object = DoubleLoopCoordinator(
        bidder=thermal_bidder,
        tracker=thermal_tracker,
        projection_tracker=thermal_projection_tracker,
    )

    return coordinator_object


@pytest.mark.unit
def test_static_params(coordinator_object):
    if (
        coordinator_object.bidder.bidding_model_object.model_data.generator_type
        == "thermal"
    ):
        gen_dict = testing_generator_params
    else:
        gen_dict = renewable_generator_params
    coordinator_object._update_static_params(gen_dict)


@pytest.mark.unit
def test_assemble_sced_tracking_market_signals(coordinator_object):

    gen_name = coordinator_object.bidder.generator
    # assumes constant sced dispatch signal in the horizon
    constant_dispatch = 20
    sced_dispatch = [constant_dispatch] * tracking_horizon

    # test case 1: no ruc signals from next day
    hour = 10
    next_ruc_dispatch_dicts = None
    coordinator_object.current_DA_dispatches = [(t + 1) * 10 for t in range(24)]
    coordinator_object.current_avail_DA_dispatches = [(t + 1) * 10 for t in range(24)]

    # expected sced signals are: the sced signal from the coming hour and
    # corresponding ruc signals (current day) for the remaining horizon
    expected_signal = [sced_dispatch[0]] + [
        (t + 1) * 10 for t in range(hour + 1, hour + tracking_horizon)
    ]
    signal = coordinator_object._assemble_sced_tracking_market_signals(
        hour=hour,
        sced_dispatch=sced_dispatch,
        tracking_horizon=tracking_horizon,
    )
    pyo_unittest.assertStructuredAlmostEqual(first=signal, second=expected_signal)

    # test case 2: no ruc signals, but between 2 days
    hour = 23
    next_ruc_dispatch_dicts = None

    # expected sced signals are: because there is no next ruc signals, the expected
    # signal will be the same as the sced dispatch
    expected_signal = [sced_dispatch[0]]
    signal = coordinator_object._assemble_sced_tracking_market_signals(
        hour=hour,
        sced_dispatch=sced_dispatch,
        tracking_horizon=tracking_horizon,
    )
    pyo_unittest.assertStructuredAlmostEqual(first=signal, second=expected_signal)

    # test case 3: with ruc signals, between 2 days
    hour = 23
    coordinator_object.current_avail_DA_dispatches += [(t + 1) * 10 for t in range(24)]

    # expected sced signals are: because there is ruc signals, the expected
    # signal will be the sced signal from the coming hour and
    # corresponding ruc signals (next day) for the remaining horizon
    expected_signal = [sced_dispatch[0]] + [
        (t + 1) * 10 for t in range(0, tracking_horizon - 1)
    ]
    signal = coordinator_object._assemble_sced_tracking_market_signals(
        hour=hour,
        sced_dispatch=sced_dispatch,
        tracking_horizon=tracking_horizon,
    )
    pyo_unittest.assertStructuredAlmostEqual(first=signal, second=expected_signal)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pandas as pd
import pyomo.environ as pyo
from pyomo.opt.base.solvers import OptSolver
import os


class Tracker:
    """
    Wrap a model object to track the market dispatch signals. This class interfaces
    with the DoubleLoopCoordinator.
    """

    def __init__(
        self, tracking_model_object, tracking_horizon, n_tracking_hour, solver
    ):
        """
        Initializes the tracker object.

        Arguments:
            tracking_model_object: the model object for tracking

            tracking_horizon: number of time periods in the tracking problem

            n_tracking_hour: number of implemented hours after each solve

            solver: a Pyomo mathematical programming solver object

        Returns:
            None
        """

        # copy and check model object
        self.tracking_model_object = tracking_model_object
        self.tracking_horizon = tracking_horizon
        self.n_tracking_hour = n_tracking_hour
        self.solver = solver
        self._check_inputs()

        # add flowsheet to model
        self.model = pyo.ConcreteModel()
        self.model.fs = pyo.Block()
        self.tracking_model_object.populate_model(self.model.fs, self.tracking_horizon)

        # get the power output
        power_output_name = self.tracking_model_object.power_output
        self.power_output = getattr(self.model.fs, power_output_name)

        # get the time index set
        self.time_set = self.power_output.index_set()

        self.formulate_tracking_problem()

        self.daily_stats = None
        self.projection = None

        self.result_list = []

    def _check_inputs(self):
        """
        Check if the inputs to construct the tracker is valid. If not raise errors.
        """

        self._check_tracking_model_object()
        self._check_n_tracking_hour()
        self._check_solver()

    def _check_tracking_model_object(self):
        """
        Check if tracking model object has the necessary methods and attributes.
        """

        method_list = [
            "populate_model",
            "get_implemented_profile",
            "update_model",
            "get_last_delivered_power",
            "record_results",
            "write_results",
        ]
        attr_list = ["power_output", "total_cost"]
        msg = "Tracking model object does not have a "

        for m in method_list:
            obtained_m = getattr(self.tracking_model_object, m, None)
            if obtained_m is None:
                raise AttributeError(
                    msg
                    + m
                    + "() method. "
                    + "The tracker object needs the users to "
                    + "implement this method in their model object."
                )

        for attr in attr_list:
            obtained_attr = getattr(self.tracking_model_object, attr, None)
            if obtained_attr is None:
                raise AttributeError(
                    msg
                    + attr
                    + " property. "
                    + "The tracker object needs the users to "
                    + "specify this property in their model object."
                )

    def _check_n_tracking_hour(self):
        """
        Check if the number of hour for tracking is an integer and greater than 0.
        """

        # check if it is an integer
        if not isinstance(self.n_tracking_hour, int):
            raise TypeError(
                "The number of hour for tracking should be an integer, "
                + "but a {} was given.".format(type(self.n_tracking_hour).__name__)
            )

        if self.n_tracking_hour <= 0:
            raise ValueError(
                "The number of hour for tracking should be greater than zero, "
                + "but {} was given.".format(self.n_tracking_hour)
            )

    def _check_solver(self):
        """
        Check if provides solver is a valid Pyomo solver object.
        """

        if not isinstance(self.solver, OptSolver):
            raise TypeError(
                "The provided solver {} is not a valid Pyomo solver.".format(
                    self.solver
                )
            )

    def formulate_tracking_problem(self):
        """
        Formulate the tracking optimization problem by adding necessary
        parameters, constraints, and objective function.

        Arguments:
            None

        Returns:
            None
        """

        self._add_tracking_params()
        self._add_tracking_vars()
        self._add_tracking_constraints()
        self._add_tracking_objective()

        return

    def _add_tracking_vars(self):
        """
        Add necessary tracking variables to the model, i.e., power under and over
        delivered.

        Arguments:
            None

        Returns:
            None
        """

        self.model.power_underdelivered = pyo.Var(
            self.time_set, initialize=0, within=pyo.NonNegativeReals
        )
        self.model.power_overdelivered = pyo.Var(
            self.time_set, initialize=0, within=pyo.NonNegativeReals
        )

        return

    def _add_tracking_params(self):
        """
        Add necessary tracking parameters to the model, i.e., market dispatch
        signal.

        Arguments:
            None

        Returns:
            None
        """

        # add params to the model
        self.model.power_dispatch = pyo.Param(
            self.time_set, initialize=0, within=pyo.Reals, mutable=True
        )

        large_penalty = 10000
        penalty_init = {}
        for t in self.time_set:
            if t < self.n_tracking_hour:
                penalty_init[t] = large_penalty
            else:
                penalty_init[t] = large_penalty / (
                    self.tracking_horizon - self.n_tracking_hour
                )
        self.model.deviation_penalty = pyo.Param(
            self.time_set, initialize=penalty_init, mutable=False
        )

        return

    def _add_tracking_constraints(self):
        """
        Add necessary tracking constraints to the model, e.g., power output needs
        to follow market dispatch signals.

        Arguments:
            None

        Returns:
            None
        """

        # declare a constraint list
        def tracking_dispatch_constraint_rule(m, t):
            return (
                self.power_output[t] + self.model.power_underdelivered[t]
                == self.model.power_dispatch[t] + self.model.power_overdelivered[t]
            )

        self.model.tracking_dispatch_constraints = pyo.Constraint(
            self.time_set, rule=tracking_dispatch_constraint_rule
        )

        return

    def _add_tracking_objective(self):
        """
        Add EMPC objective function to the model, i.e., minimizing different costs
        of the energy system.

        Arguments:
            None

        Returns:
            None
        """

        # declare an empty objective
        self.model.obj = pyo.Objective(expr=0, sense=pyo.minimize)

        cost_name = self.tracking_model_object.total_cost[0]
        cost = getattr(self.model.fs, cost_name)
        weight = self.tracking_model_object.total_cost[1]

        for t in self.time_set:
            self.model.obj.expr += weight * cost[t] + self.model.deviation_penalty[
                t
            ] * (self.model.power_underdelivered[t] + self.model.power_overdelivered[t])

        return

    def update_model(self, **profiles):
        """
        This method updates the parameters in the model based on the implemented profiles.

        Arguments:
            profiles: the newly implemented stats. {stat_name: [...]}

        Returns:
            None
        """

        self.tracking_model_object.update_model(self.model.fs, **profiles)

    def track_market_dispatch(self, market_dispatch, date, hour):
        """
        Solve the model to track the market dispatch signals. After solving,
        record the results from the solve and update the model.

        Arguments:
            market_dispatch: a list that contains the market dispatch signals

            date: current simulation date

            hour: current simulation hour

        Returns:
            None
        """

        self._pass_market_dispatch(market_dispatch)

        # solve the model
        self.solver.solve(self.model, tee=False)

        self.record_results(date=date, hour=hour)

        # update the model
        profiles = self.tracking_model_object.get_implemented_profile(
            b=self.model.fs, last_implemented_time_step=self.n_tracking_hour - 1
        )

        self._record_daily_stats(profiles)

        return profiles

    def _record_daily_stats(self, profiles):
        """
        Record the stats that are used to update the model in the past 24 hours.

        Arguments:
            profiles: the newly implemented stats. {stat_name: [...]}

        Returns:
            None
        """

        if self.daily_stats is None:
            self.daily_stats = profiles
        else:
            for k in self.daily_stats:
                self.daily_stats[k] += profiles[k]

        for v in self.daily_stats.values():
            while len(v) >= 24:
                v.popleft()

        return

    def _pass_market_dispatch(self, market_dispatch):
        """
        Pass the received market signals into model parameters.

        Arguments:
            market_dispatch: a list that contains the market dispatch signals

        Returns:
            None
        """

        for t in self.time_set:

            try:
                dispatch = market_dispatch[t]
            except IndexError:
                self.model.tracking_dispatch_constraints[t].deactivate()
            else:
                self.model.power_dispatch[t] = dispatch
                self.model.tracking_dispatch_constraints[t].activate()

        return

    def get_last_delivered_power(self):
        """
        Returns the last delivered power output.

        Arguments:
            None

        Returns:
            None
        """
        return self.tracking_model_object.get_last_delivered_power(
            b=self.model.fs, last_implemented_time_step=self.n_tracking_hour - 1
        )

    def _record_tracker_results(self, **kwargs):
        """
        Record the tracker stats.

        Arguments:
            kwargs: key word arguments that can be passed into tracking model object's record result function.

        Returns:
            None

        """

        df_list = []
        for t in self.time_set:

            result_dict = {}

            result_dict["Date"] = kwargs["date"]
            result_dict["Hour"] = kwargs["hour"]

            result_dict["Horizon [hr]"] = int(t)

            result_dict["Power Dispatch [MW]"] = round(
                pyo.value(self.model.power_dispatch[t]), 2
            )
            result_dict["Power Output [MW]"] = round(pyo.value(self.power_output[t]), 2)
            result_dict["Power Underdelivered [MW]"] = round(
                pyo.value(self.model.power_underdelivered[t]), 2
            )
            result_dict["Power Overdelivered [MW]"] = round(
                pyo.value(self.model.power_overdelivered[t]), 2
            )

            result_df = pd.DataFrame.from_dict(result_dict, orient="index")
            df_list.append(result_df.T)

        # append to result list
        self.result_list.append(pd.concat(df_list))

    def record_results(self, **kwargs):
        """
        Record the operations stats for the model.

        Arguments:
            kwargs: key word arguments that can be passed into tracking model object's record result function.

        Returns:
            None

        """

        # record tracker details
        self._record_tracker_results(**kwargs)

        # tracking model details
        self.tracking_model_object.record_results(self.model.fs, **kwargs)

    def write_results(self, path):
        """
        This methods writes the saved operation stats into an csv file.

        Arguments:
            path: the path to write the results.

        Return:
            None
        """

        print("")
        print("Saving tracking results to disk...")

        pd.concat(self.result_list).to_csv(
            os.path.join(path, "tracker_detail.csv"), index=False
        )
        self.tracking_model_object.write_results(
            path=os.path.join(path, "tracking_model_detail.csv")
        )
